using System;
using System.Diagnostics;
using System.Collections.Generic;
using OpenTK;
using OpenTK.Graphics;
using OpenTK.Graphics.OpenGL;

public class GridGeometry
{
    public int textureId;
    public int selectionTextureId;
    public int puffTextureId;
    public Shader shader = new Shader();
    private uint puffVboId;
    private uint gridVboId;
    private uint selection1VboId;
    private int vertexCount;
    private int selectionVertexCount;
    private Vector4 topLeftCorner;
    private float blockWidth;
    private float blockSpacing;

    public class Puff
    {
        public static int lifeTimeMs = 1100;

        public Puff(int r, int c, long spawnTime)
        {
            row = r;
            column = c;
            spawnTimeMs = spawnTime;
        }

        public int row;
        public int column;
        public long spawnTimeMs;
    }

    private Vector2 GetOffset( int row, int column )
    {
        return new Vector2((float)column * (float)(blockWidth + blockSpacing) + topLeftCorner.X, (float)row * (float)(blockWidth + blockSpacing) + topLeftCorner.Y);
    }


    public void CreateSelection1( int row, int column )
    {
        GL.DeleteBuffer(selection1VboId);

        Vector4 offset = new Vector4((float)column * (float)(blockWidth + blockSpacing), (float)row * (float)(blockWidth + blockSpacing), 0, 0);

        // UV is stored in .zw
        selectionVertexCount = 6;
        Vector4[] vertices = new Vector4[ vertexCount ];

        float u0 = 1;
        float u1 = 1;

        // Upper triangle.
        vertices[0] = new Vector4(0, 0, u0, 1) + topLeftCorner + offset;
        vertices[1] = new Vector4(blockWidth, 0, u1, 1) + topLeftCorner + offset;
        vertices[2] = new Vector4(blockWidth, blockWidth, u1, 0) + topLeftCorner + offset;

        // Lower triangle.
        vertices[3] = new Vector4(0, 0, u0, 1) + topLeftCorner + offset;
        vertices[4] = new Vector4(blockWidth, blockWidth, u1, 0) + topLeftCorner + offset;
        vertices[5] = new Vector4(0, blockWidth, u0, 0) + topLeftCorner + offset;

        GL.GenBuffers(1, out selection1VboId);
        GL.BindBuffer(BufferTarget.ArrayBuffer, selection1VboId);
        GL.BufferData< Vector4 >(BufferTarget.ArrayBuffer, new IntPtr(vertices.Length * Vector4.SizeInBytes), vertices, BufferUsageHint.StaticDraw);

        GL.EnableVertexAttribArray(0);
        GL.VertexAttribPointer(0, 4, VertexAttribPointerType.Float, false, Vector4.SizeInBytes, 0);
        GL.BindBuffer(BufferTarget.ArrayBuffer, 0);
    }

    // Must be called after Create().
    public void InitSelectionGraphicsObjects()
    {
        Debug.Assert(blockWidth > 0);

        // UV is stored in .zw
        selectionVertexCount = 4;
        Vector4[] vertices = new Vector4[ vertexCount ];
        vertices[0] = new Vector4(0, 0, 0, 0) + topLeftCorner;
        vertices[1] = new Vector4(blockWidth, 0, 1, 0) + topLeftCorner;
        vertices[2] = new Vector4(blockWidth, blockWidth, 1, 1) + topLeftCorner;
        vertices[3] = new Vector4(0, blockWidth, 0, 1) + topLeftCorner;

        GL.GenBuffers(1, out selection1VboId);
        GL.BindBuffer(BufferTarget.ArrayBuffer, selection1VboId);
        GL.BufferData< Vector4 >(BufferTarget.ArrayBuffer, new IntPtr(vertices.Length * Vector4.SizeInBytes), vertices, BufferUsageHint.StaticDraw);

        GL.EnableVertexAttribArray(0);
        GL.VertexAttribPointer(0, 4, VertexAttribPointerType.Float, false, Vector4.SizeInBytes, 0);
        GL.BindBuffer(BufferTarget.ArrayBuffer, 0);

        CreatePuffVbo();
    }


    public void Create( int[,] grid, float spriteWidth, float spacing, Vector4 offset )
    {
        blockWidth = spriteWidth;
        blockSpacing = spacing;
        topLeftCorner = offset;

        vertexCount = grid.GetLength(0) * grid.GetLength(1) * 2 * 3;

        // UV is stored in .zw
        Vector4[] vertices = new Vector4[ vertexCount ];

        int i = 0;

        for (int row = 0; row < grid.GetLength(0); ++row)
        {
            for (int col = 0; col < grid.GetLength(1); ++col)
            {
                float u0 = (1.0f / GameState.blockTypeCount) * grid[row, col];
                float u1 = u0 + (1.0f / GameState.blockTypeCount);

                // Upper triangle.
                vertices[i * 6 + 0] = new Vector4(col * (spriteWidth + spacing), row * (spriteWidth + spacing), u0, 1) + offset;
                vertices[i * 6 + 1] = new Vector4(spriteWidth + col * (spriteWidth + spacing), row * (spriteWidth + spacing), u1, 1) + offset;
                vertices[i * 6 + 2] = new Vector4(spriteWidth + col * (spriteWidth + spacing), spriteWidth + row * (spriteWidth + spacing), u1, 0) + offset;
			
                // Lower triangle.
                vertices[i * 6 + 3] = new Vector4(col * (spriteWidth + spacing), row * (spriteWidth + spacing), u0, 1) + offset;
                vertices[i * 6 + 4] = new Vector4(spriteWidth + col * (spriteWidth + spacing), spriteWidth + row * (spriteWidth + spacing), u1, 0) + offset;
                vertices[i * 6 + 5] = new Vector4(col * (spriteWidth + spacing), spriteWidth + row * (spriteWidth + spacing), u0, 0) + offset;

                ++i;
            }
        }

        GL.GenBuffers(1, out gridVboId);
        GL.BindBuffer(BufferTarget.ArrayBuffer, gridVboId);
        GL.BufferData< Vector4 >(BufferTarget.ArrayBuffer, new IntPtr(vertices.Length * Vector4.SizeInBytes), vertices, BufferUsageHint.StaticDraw);

        GL.EnableVertexAttribArray(0);
        GL.VertexAttribPointer(0, 4, VertexAttribPointerType.Float, false, Vector4.SizeInBytes, 0);
        GL.BindBuffer(BufferTarget.ArrayBuffer, 0);
    }


    public void Draw()
    {
        GL.BindTexture(TextureTarget.Texture2D, textureId);
        GL.BindBuffer(BufferTarget.ArrayBuffer, gridVboId);
        GL.EnableVertexAttribArray(0);
        GL.VertexAttribPointer(0, 4, VertexAttribPointerType.Float, false, Vector4.SizeInBytes, 0);

        GL.DrawArrays(PrimitiveType.Triangles, 0, vertexCount);
    }


    public void DrawSelection1( int row, int column )
    {
        GL.BindTexture(TextureTarget.Texture2D, selectionTextureId);
        GL.BindBuffer(BufferTarget.ArrayBuffer, selection1VboId);
        GL.EnableVertexAttribArray(0);
        GL.VertexAttribPointer(0, 4, VertexAttribPointerType.Float, false, Vector4.SizeInBytes, 0);

        GL.Enable(EnableCap.Blend);
        GL.BlendEquation(BlendEquationMode.FuncAdd);
        GL.DrawArrays(PrimitiveType.Triangles, 0, selectionVertexCount);
        GL.Disable(EnableCap.Blend);
    }


    public void GetBlockCoordinate( int screenX, int screenY, out int selectedRow, out int selectedColumn )
    {
        selectedRow = (int)(((int)-topLeftCorner.Y - screenY + Game.screenSize.Y / (3 - Game.ScaleFactor)) / (blockWidth + blockSpacing));
        selectedColumn = (int)(((int)topLeftCorner.X - screenX + Game.screenSize.X / (3 - Game.ScaleFactor)) / -(blockWidth + blockSpacing));
    }


    public void DrawPuffs( List< Puff > puffs )
    {
        foreach (var puff in puffs)
        {
            DrawPuff(puff.row, puff.column, 1 - (MatchFree.Utils.stopWatch.ElapsedMilliseconds - puff.spawnTimeMs) / (float)Puff.lifeTimeMs);
        }
    }


    private void DrawPuff( int row, int column, float opacity )
    {
        GL.BindTexture(TextureTarget.Texture2D, puffTextureId);
        GL.BindBuffer(BufferTarget.ArrayBuffer, puffVboId);
        GL.EnableVertexAttribArray(0);
        GL.VertexAttribPointer(0, 4, VertexAttribPointerType.Float, false, Vector4.SizeInBytes, 0);

        shader.SetVector(GetOffset(row, column), Shader.Uniform.Translation);
        shader.SetFloat(opacity, Shader.Uniform.Opacity);
        GL.Enable(EnableCap.Blend);
        GL.BlendFunc(BlendingFactorSrc.SrcAlpha, BlendingFactorDest.OneMinusSrcAlpha);
        GL.DrawArrays(PrimitiveType.Triangles, 0, selectionVertexCount);
        GL.Disable(EnableCap.Blend);
        shader.SetVector(Vector2.Zero, Shader.Uniform.Translation);
        shader.SetFloat(1.0f, Shader.Uniform.Opacity);
    }


    private void CreatePuffVbo()
    {
        GL.DeleteBuffer(puffVboId);

        // UV is stored in .zw
        int vertexCount = 6;
        Vector4[] vertices = new Vector4[ vertexCount ];

        float u0 = 0;
        float u1 = 1;

        // Upper triangle.
        vertices[0] = new Vector4(0, 0, u0, 1);
        vertices[1] = new Vector4(blockWidth, 0, u1, 1);
        vertices[2] = new Vector4(blockWidth, blockWidth, u1, 0);

        // Lower triangle.
        vertices[3] = new Vector4(0, 0, u0, 1);
        vertices[4] = new Vector4(blockWidth, blockWidth, u1, 0);
        vertices[5] = new Vector4(0, blockWidth, u0, 0);

        GL.GenBuffers(1, out puffVboId);
        GL.BindBuffer(BufferTarget.ArrayBuffer, puffVboId);
        GL.BufferData< Vector4 >(BufferTarget.ArrayBuffer, new IntPtr(vertices.Length * Vector4.SizeInBytes), vertices, BufferUsageHint.StaticDraw);

        GL.EnableVertexAttribArray(0);
        GL.VertexAttribPointer(0, 4, VertexAttribPointerType.Float, false, Vector4.SizeInBytes, 0);
        GL.BindBuffer(BufferTarget.ArrayBuffer, 0);
    }
}

